import re
import binascii
import asyncio
import logging
from typing import DefaultDict
from asyncio.coroutines import coroutine
from asyncio.futures import Future
from asyncio.transports import DatagramTransport
from aioupnp.fault import UPnPError
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.constants import UPNP_ORG_IGD, SSDP_IP_ADDRESS, SSDP_PORT, SSDP_DISCOVER, SSDP_ROOT_DEVICE
from aioupnp.protocols.multicast import MulticastProtocol

ADDRESS_REGEX = re.compile("^http:\/\/(\d+\.\d+\.\d+\.\d+)\:(\d*)(\/[\w|\/|\:|\-|\.]*)$")

log = logging.getLogger(__name__)


class SSDPProtocol(MulticastProtocol):
    def __init__(self, lan_address):
        super().__init__()
        self.lan_address = lan_address
        self.discover_callbacks: DefaultDict[coroutine] = {}
        self.transport: DatagramTransport
        self.notifications = []
        self.replies = []

    def connection_made(self, transport: DatagramTransport):
        super().connection_made(transport)
        self.set_ttl(1)

    async def m_search(self, address, timeout: int = 1, service=UPNP_ORG_IGD) -> SSDPDatagram:
        if (address, service) in self.discover_callbacks:
            return self.discover_callbacks[(address, service)]
        packet = SSDPDatagram(
            SSDPDatagram._M_SEARCH, host="{}:{}".format(SSDP_IP_ADDRESS, SSDP_PORT), st=service, man=SSDP_DISCOVER,
            mx=1
        )
        self.transport.sendto(packet.encode().encode(), (address, SSDP_PORT))
        f = Future()
        self.discover_callbacks[(address, service)] = f
        return await asyncio.wait_for(f, timeout)

    def datagram_received(self, data, addr) -> None:
        if addr[0] == self.lan_address:
            return
        try:
            packet = SSDPDatagram.decode(data)
            log.debug("decoded %s from %s:%i:\n%s", packet.get_friendly_name(), addr[0], addr[1], packet.encode())
        except UPnPError as err:
            log.error("failed to decode SSDP packet from %s:%i: %s\npacket: %s", addr[0], addr[1], err,
                      binascii.hexlify(data))
            return

        if packet._packet_type == packet._OK:
            log.debug("%s:%i sent us an OK", addr[0], addr[1])
            if (addr[0], packet.st) in self.discover_callbacks:
                if packet.st not in map(lambda p: p['st'], self.replies):
                    self.replies.append(packet)
                f: Future = self.discover_callbacks.pop((addr[0], packet.st))
                f.set_result(packet)

        elif packet._packet_type == packet._NOTIFY:
            if packet.nt == SSDP_ROOT_DEVICE:
                address, port, path = ADDRESS_REGEX.findall(packet.location)[0]
                key = None
                for (addr, service) in self.discover_callbacks:
                    if addr == address:
                        key = (addr, service)
                        break
                if key:
                    log.debug("got a notification with the requested m-search info")
                    f: Future = self.discover_callbacks.pop(key)
                    f.set_result(SSDPDatagram(
                        SSDPDatagram._OK, cache_control='', location=packet.location, server=packet.server,
                        st=UPNP_ORG_IGD, usn=packet.usn
                    ))
                self.notifications.append(packet.as_dict())


async def listen_ssdp(lan_address: str, gateway_address: str) -> (DatagramTransport, SSDPProtocol, str, str):
    loop = asyncio.get_running_loop()
    try:
        sock = SSDPProtocol.create_socket(lan_address, SSDP_IP_ADDRESS)
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: SSDPProtocol(lan_address), sock=sock
        )
    except Exception:
        log.exception("failed to create multicast socket %s:%i", lan_address, SSDP_PORT)
        raise
    return transport, protocol, gateway_address, lan_address


async def m_search(lan_address: str, gateway_address: str, timeout: int = 1,
                   service: str = UPNP_ORG_IGD) -> SSDPDatagram:
    transport, protocol, gateway_address, lan_address = await listen_ssdp(
        lan_address, gateway_address
    )
    try:
        return await protocol.m_search(gateway_address, timeout=timeout, service=service)
    except asyncio.TimeoutError:
        raise UPnPError("M-SEARCH for {}:{} timed out".format(gateway_address, SSDP_PORT))
    finally:
        transport.close()



import logging
from xml.etree import ElementTree
import asyncio
from asyncio.protocols import Protocol
from aioupnp.fault import UPnPError
from aioupnp.serialization.scpd import deserialize_scpd_get_response
from aioupnp.serialization.scpd import serialize_scpd_get
from aioupnp.serialization.soap import serialize_soap_post, deserialize_soap_post_response


log = logging.getLogger(__name__)


class SCPDHTTPClientProtocol(Protocol):
    POST = 'POST'
    GET = 'GET'

    def __init__(self, method: str, message: bytes, finished: asyncio.Future, soap_method: str=None,
                 soap_service_id: str=None, close_after_send: bool = False):
        self.method = method
        assert soap_service_id is not None and soap_method is not None if method == 'POST' else True, \
            'soap args not provided'
        self.message = message
        self.response_buff = b""
        self.finished = finished
        self.soap_method = soap_method
        self.soap_service_id = soap_service_id
        self.close_after_send = close_after_send

    def connection_made(self, transport):
        transport.write(self.message)
        if self.close_after_send:
            self.finished.set_result(None)

    def data_received(self, data):
        self.response_buff += data
        if self.method == self.GET:
            try:
                packet = deserialize_scpd_get_response(self.response_buff)
                if not packet:
                    return
            except ElementTree.ParseError:
                pass
            except UPnPError as err:
                self.finished.set_exception(err)
            else:
                self.finished.set_result(packet)
        elif self.method == self.POST:
            try:
                packet = deserialize_soap_post_response(self.response_buff, self.soap_method, self.soap_service_id)
                if not packet:
                    return
            except ElementTree.ParseError:
                pass
            except UPnPError as err:
                self.finished.set_exception(err)
            else:
                self.finished.set_result(packet)


async def scpd_get(control_url: str, address: str, port: int) -> dict:
    loop = asyncio.get_running_loop()
    finished = asyncio.Future()
    packet = serialize_scpd_get(control_url, address)
    transport, protocol = await loop.create_connection(
        lambda : SCPDHTTPClientProtocol('GET', packet, finished),  address, port
    )
    try:
        return await asyncio.wait_for(finished, 1.0)
    finally:
        transport.close()


async def scpd_post(control_url: str, address: str, port: int, method: str, param_names: list, service_id: bytes,
                    close_after_send: bool, **kwargs):
    loop = asyncio.get_running_loop()
    finished = asyncio.Future()
    packet = serialize_soap_post(method, param_names, service_id, address.encode(), control_url.encode(), **kwargs)
    transport, protocol = await loop.create_connection(
        lambda : SCPDHTTPClientProtocol(
            'POST', packet, finished, soap_method=method, soap_service_id=service_id.decode(),
            close_after_send=close_after_send
        ), address, port
    )
    try:
        return await asyncio.wait_for(finished, 1.0)
    finally:
        transport.close()



import logging
import sys
from aioupnp.upnp import UPnP
from aioupnp.constants import UPNP_ORG_IGD


log = logging.getLogger("aioupnp")
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)-15s-%(filename)s:%(lineno)s->%(message)s'))
log.addHandler(handler)
log.setLevel(logging.WARNING)


def get_help(command):
    fn = getattr(UPnP, command)
    params = command + " " + " ".join(["[--%s=<%s>]" % (k, k) for k in fn.__annotations__ if k != 'return'])
    return \
        "usage: aioupnp [--debug_logging=<debug_logging>] [--interface=<interface>]\n" \
        "              [--gateway_address=<gateway_address>]\n" \
        "              [--lan_address=<lan_address>] [--timeout=<timeout>]\n" \
        "              [--service=<service>]\n" \
        "              %s\n" % params


def main():
    commands = [n for n in dir(UPnP) if hasattr(getattr(UPnP, n, None), "_cli")]
    help_str = " | ".join(commands)
    usage = \
        "usage: aioupnp [-h] [--debug_logging=<debug_logging>] [--interface=<interface>]\n" \
        "              [--gateway_address=<gateway_address>]\n" \
        "              [--lan_address=<lan_address>] [--timeout=<timeout>]\n" \
        "              [--service=<service>]\n" \
        "              command [--<arg name>=<arg>]...\n" \
        "\n" \
        "commands: %s\n\nfor help with a specific command: aioupnp help <command>" % help_str

    args = sys.argv[1:]
    if args[0] in ['help', '-h', '--help']:
        if len(args) > 1:
            if args[1] in commands:
                sys.exit(get_help(args[1]))
        sys.exit(print(usage))

    defaults = {
        'debug_logging': False,
        'interface': 'default',
        'gateway_address': '',
        'lan_address': '',
        'timeout': 1,
        'service': UPNP_ORG_IGD,
        'return_as_json': True
    }

    options = {}
    command = None
    for arg in args:
        if arg.startswith("--"):
            k, v = arg.split("=")
            k = k.lstrip('--')
            options[k] = v
        else:
            command = arg
            break
    if not command:
        print("no command given")
        sys.exit(print(usage))
    kwargs = {}
    for arg in args[len(options)+1:]:
        if arg.startswith("--"):
            k, v = arg.split("=")
            k = k.lstrip('--')
            kwargs[k] = v
        else:
            break
    for k, v in defaults.items():
        if k not in options:
            options[k] = v

    if options.pop('debug_logging'):
        log.setLevel(logging.DEBUG)

    UPnP.run_cli(
        command.replace('-', '_'), options.pop('lan_address'), options.pop('gateway_address'),
        options.pop('timeout'), options.pop('service'), options.pop('interface'),
        kwargs
    )


if __name__ == "__main__":
    main()



import logging
from aioupnp.util import get_dict_val_case_insensitive, BASE_PORT_REGEX, BASE_ADDRESS_REGEX
from aioupnp.constants import SPEC_VERSION, UPNP_ORG_IGD, SERVICE
from aioupnp.commands import SCPDCommands
from aioupnp.device import Device, Service
from aioupnp.protocols.ssdp import m_search
from aioupnp.protocols.scpd import scpd_get
from aioupnp.protocols.soap import SCPDCommand
from aioupnp.util import flatten_keys

log = logging.getLogger(__name__)


def get_action_list(element_dict: dict) -> list:  # [(<method>, [<input1>, ...], [<output1, ...]), ...]
    service_info = flatten_keys(element_dict, "{%s}" % SERVICE)
    if "actionList" in service_info:
        action_list = service_info["actionList"]
    else:
        return []
    if not len(action_list):  # it could be an empty string
        return []

    result = []
    if isinstance(action_list["action"], dict):
        arg_dicts = action_list["action"]['argumentList']['argument']
        if not isinstance(arg_dicts, list):  # when there is one arg
            arg_dicts = [arg_dicts]
        return [[
            action_list["action"]['name'],
            [i['name'] for i in arg_dicts if i['direction'] == 'in'],
            [i['name'] for i in arg_dicts if i['direction'] == 'out']
        ]]
    for action in action_list["action"]:
        if not action.get('argumentList'):
            result.append((action['name'], [], []))
        else:
            arg_dicts = action['argumentList']['argument']
            if not isinstance(arg_dicts, list):  # when there is one arg
                arg_dicts = [arg_dicts]
            result.append((
                action['name'],
                [i['name'] for i in arg_dicts if i['direction'] == 'in'],
                [i['name'] for i in arg_dicts if i['direction'] == 'out']
            ))
    return result


class Gateway:
    def __init__(self, **kwargs):
        flattened = {
            k.lower(): v for k, v in kwargs.items()
        }
        usn = flattened["usn"]
        server = flattened["server"]
        location = flattened["location"]
        st = flattened["st"]

        cache_control = flattened.get("cache_control") or flattened.get("cache-control") or ""
        date = flattened.get("date", "")
        ext = flattened.get("ext", "")

        self.usn = usn.encode()
        self.ext = ext.encode()
        self.server = server.encode()
        self.location = location.encode()
        self.cache_control = cache_control.encode()
        self.date = date.encode()
        self.urn = st.encode()

        self._xml_response = ""
        self._service_descriptors = {}
        self.base_address = BASE_ADDRESS_REGEX.findall(self.location)[0]
        self.port = int(BASE_PORT_REGEX.findall(self.location)[0])
        self.base_ip = self.base_address.lstrip(b"http://").split(b":")[0]
        self.path = self.location.split(b"%s:%i/" % (self.base_ip, self.port))[1]

        self.spec_version = None
        self.url_base = None

        self._device = None
        self._devices = []
        self._services = []

        self._unsupported_actions = {}
        self._registered_commands = {}
        self.commands = SCPDCommands()

    def gateway_descriptor(self) -> dict:
        r = {
            'server': self.server.decode(),
            'urlBase': self.url_base,
            'location': self.location.decode(),
            "specVersion": self.spec_version,
            'usn': self.usn.decode(),
            'urn': self.urn.decode(),
        }
        return r

    @property
    def services(self) -> dict:
        if not self._device:
            return {}
        return {service.serviceType: service for service in self._services}

    @property
    def devices(self) -> dict:
        if not self._device:
            return {}
        return {device.udn: device for device in self._devices}

    def get_service(self, service_type: str) -> Service:
        for service in self._services:
            if service.serviceType.lower() == service_type.lower():
                return service

    def debug_commands(self):
        return {
            'available': self._registered_commands,
            'failed': self._unsupported_actions
        }

    @classmethod
    async def discover_gateway(cls, lan_address: str, gateway_address: str, timeout: int = 1,
                               service: str = UPNP_ORG_IGD):
        datagram = await m_search(lan_address, gateway_address, timeout, service)
        gateway = cls(**datagram.as_dict())
        await gateway.discover_commands()
        return gateway

    async def discover_commands(self):
        response = await scpd_get("/" + self.path.decode(), self.base_ip.decode(), self.port)

        self.spec_version = get_dict_val_case_insensitive(response, SPEC_VERSION)
        self.url_base = get_dict_val_case_insensitive(response, "urlbase")
        if not self.url_base:
            self.url_base = self.base_address.decode()
        if response:
            self._device = Device(
                self._devices, self._services, **get_dict_val_case_insensitive(response, "device")
            )
        else:
            self._device = Device(self._devices, self._services)
        for service_type in self.services.keys():
            await self.register_commands(self.services[service_type])

    async def register_commands(self, service: Service):
        service_dict = await scpd_get(("" if service.SCPDURL.startswith("/") else "/") + service.SCPDURL,
                                      self.base_ip.decode(), self.port)
        if not service_dict:
            return

        action_list = get_action_list(service_dict)

        for name, inputs, outputs in action_list:
            try:
                current = getattr(self.commands, name)
                annotations = current.__annotations__
                return_types = annotations.get('return', None)
                if return_types:
                    if not isinstance(return_types, tuple):
                        return_types = (return_types, )
                    return_types = {r: t for r, t in zip(outputs, return_types)}
                param_types = {}
                for param_name, param_type in annotations.items():
                    if param_name == "return":
                        continue
                    param_types[param_name] = param_type
                command = SCPDCommand(
                    self.base_ip.decode(), self.port, service.controlURL, service.serviceType.encode(),
                    name, param_types, return_types, inputs, outputs)
                setattr(command, "__doc__", current.__doc__)
                setattr(self.commands, command.method, command)

                self._registered_commands[command.method] = service.serviceType
                log.debug("registered %s::%s", service.serviceType, command.method)
            except AttributeError:
                s = self._unsupported_actions.get(service.serviceType, [])
                s.append(name)
                self._unsupported_actions[service.serviceType] = s
                log.debug("available command for %s does not have a wrapper implemented: %s %s %s",
                          service.serviceType, name, inputs, outputs)



import logging
from aioupnp.protocols.scpd import scpd_post

log = logging.getLogger(__name__)


class SCPDCommand:
    def __init__(self, gateway_address: str, service_port: int, control_url: str, service_id: bytes, method: str,
                 param_types: dict, return_types: dict, param_order: list, return_order: list):
        self.gateway_address = gateway_address
        self.service_port = service_port
        self.control_url = control_url
        self.service_id = service_id
        self.method = method
        self.param_types = param_types
        self.param_order = param_order
        self.return_types = return_types
        self.return_order = return_order

    async def __call__(self, **kwargs):
        if set(kwargs.keys()) != set(self.param_types.keys()):
            raise Exception("argument mismatch: %s vs %s" % (kwargs.keys(), self.param_types.keys()))
        close_after_send = not self.return_types or self.return_types == [None]
        response = await scpd_post(
            self.control_url, self.gateway_address, self.service_port, self.method, self.param_order, self.service_id,
            close_after_send, **{n: self.param_types[n](kwargs[n]) for n in self.param_types.keys()}
        )
        extracted_response = tuple([None if self.return_types[n] is None else self.return_types[n](response[n])
                                    for n in self.return_order]) or (None, )
        if len(extracted_response) == 1:
            return extracted_response[0]
        return extracted_response



import os
import logging
import json
import asyncio
import functools
from aioupnp.fault import UPnPError
from aioupnp.gateway import Gateway
from aioupnp.constants import UPNP_ORG_IGD
from aioupnp.util import get_gateway_and_lan_addresses
from aioupnp.protocols.ssdp import m_search

log = logging.getLogger(__name__)


def cli(format_result=None):
    def _cli(fn):
        @functools.wraps(fn)
        async def _inner(*args, **kwargs):
            result = await fn(*args, **kwargs)
            if not format_result or not result or not isinstance(result, (list, dict, tuple)):
                return result
            self = args[0]
            return {k: v for k, v in zip(getattr(self.gateway.commands, format_result).return_order, result)}
        f = _inner
        f._cli = True
        return f
    return _cli


def _encode(x):
    if isinstance(x, bytes):
        return x.decode()
    elif isinstance(x, Exception):
        return str(x)
    return x


class UPnP:
    def __init__(self, lan_address: str, gateway_address: str, gateway: Gateway):
        self.lan_address = lan_address
        self.gateway_address = gateway_address
        self.gateway = gateway

    @classmethod
    def get_lan_and_gateway(cls, lan_address: str = '', gateway_address: str = '',
                            interface_name: str = 'default') -> (str, str):
        if not lan_address or not gateway_address:
            gateway_addr, lan_addr = get_gateway_and_lan_addresses(interface_name)
            lan_address = lan_address or lan_addr
            gateway_address = gateway_address or gateway_addr
        return lan_address, gateway_address

    @classmethod
    async def discover(cls, lan_address: str = '', gateway_address: str = '', timeout: int = 1,
                       service: str = UPNP_ORG_IGD, interface_name: str = 'default'):
        try:
            lan_address, gateway_address = cls.get_lan_and_gateway(lan_address, gateway_address, interface_name)
        except Exception as err:
            raise UPnPError("failed to get lan and gateway addresses: %s" % str(err))
        gateway = await Gateway.discover_gateway(lan_address, gateway_address, timeout, service)
        return cls(lan_address, gateway_address, gateway)

    @classmethod
    @cli()
    async def m_search(cls, lan_address: str = '', gateway_address: str = '', timeout: int = 1,
                       service: str = UPNP_ORG_IGD, interface_name: str = 'default') -> dict:
        lan_address, gateway_address = cls.get_lan_and_gateway(lan_address, gateway_address, interface_name)
        datagram = await m_search(lan_address, gateway_address, timeout, service)
        return {
            'lan_address': lan_address,
            'gateway_address': gateway_address,
            'discover_reply': datagram.as_dict()
        }

    @cli()
    async def get_external_ip(self) -> str:
        return await self.gateway.commands.GetExternalIPAddress()

    @cli("AddPortMapping")
    async def add_port_mapping(self, external_port: int, protocol: str, internal_port, lan_address: str,
                         description: str) -> None:
        return await self.gateway.commands.AddPortMapping(
            NewRemoteHost="", NewExternalPort=external_port, NewProtocol=protocol,
            NewInternalPort=internal_port, NewInternalClient=lan_address,
            NewEnabled=1, NewPortMappingDescription=description, NewLeaseDuration=""
        )

    @cli("GetGenericPortMappingEntry")
    async def get_port_mapping_by_index(self, index: int) -> dict:
        return await self._get_port_mapping_by_index(index)

    async def _get_port_mapping_by_index(self, index: int) -> (str, int, str, int, str, bool, str, int):
        try:
            redirect = await self.gateway.commands.GetGenericPortMappingEntry(NewPortMappingIndex=index)
            return redirect
        except UPnPError:
            return

    @cli()
    async def get_redirects(self) -> list:
        redirects = []
        cnt = 0
        redirect = await self.get_port_mapping_by_index(cnt)
        while redirect:
            redirects.append(redirect)
            cnt += 1
            redirect = await self.get_port_mapping_by_index(cnt)
        return redirects

    @cli("GetSpecificPortMappingEntry")
    async def get_specific_port_mapping(self, external_port: int, protocol: str):
        """
        :param external_port: (int) external port to listen on
        :param protocol:      (str) 'UDP' | 'TCP'
        :return: (int) <internal port>, (str) <lan ip>, (bool) <enabled>, (str) <description>, (int) <lease time>
        """

        try:
            return await self.gateway.commands.GetSpecificPortMappingEntry(
                NewRemoteHost=None, NewExternalPort=external_port, NewProtocol=protocol
            )
        except UPnPError:
            return

    @cli()
    async def delete_port_mapping(self, external_port: int, protocol: str) -> None:
        """
        :param external_port: (int) external port to listen on
        :param protocol:      (str) 'UDP' | 'TCP'
        :return: None
        """
        return await self.gateway.commands.DeletePortMapping(
            NewRemoteHost="", NewExternalPort=external_port, NewProtocol=protocol
        )

    @cli("AddPortMapping")
    async def get_next_mapping(self, port: int, protocol: str, description: str, internal_port: int=None) -> int:
        if protocol not in ["UDP", "TCP"]:
            raise UPnPError("unsupported protocol: {}".format(protocol))
        internal_port = internal_port or port
        redirect_tups = []
        cnt = 0
        port = int(port)
        internal_port = int(internal_port)
        redirect = await self._get_port_mapping_by_index(cnt)
        while redirect:
            redirect_tups.append(redirect)
            cnt += 1
            redirect = await self._get_port_mapping_by_index(cnt)

        redirects = {
            "%i:%s" % (ext_port, proto): (int_host, int_port, desc)
            for (ext_host, ext_port, proto, int_port, int_host, enabled, desc, lease) in redirect_tups
        }
        while ("%i:%s" % (port, protocol)) in redirects:
            int_host, int_port, _ = redirects["%i:%s" % (port, protocol)]
            if int_host == self.lan_address and int_port == internal_port:
                break
            port += 1

        await self.add_port_mapping(  # set one up
                port, protocol, internal_port, self.lan_address, description
        )
        return port

    @cli()
    async def get_soap_commands(self) -> dict:
        return {
            'supported': list(self.gateway._registered_commands.keys()),
            'unsupported': self.gateway._unsupported_actions
        }

    @cli()
    async def generate_test_data(self):
        external_ip = await self.get_external_ip()
        redirects = await self.get_redirects()
        ext_port = await self.get_next_mapping(4567, "UDP", "aioupnp test mapping")
        delete = await self.delete_port_mapping(ext_port, "UDP")
        after_delete = await self.get_specific_port_mapping(ext_port, "UDP")

        commands_test_case = (
            ("get_external_ip", (), "1.2.3.4"),
            ("get_redirects", (), redirects),
            ("get_next_mapping", (4567, "UDP", "aioupnp test mapping"), ext_port),
            ("delete_port_mapping", (ext_port, "UDP"), delete),
            ("get_specific_port_mapping", (ext_port, "UDP"), after_delete),
        )

        gateway = self.gateway
        device = list(gateway.devices.values())[0]
        assert device.manufacturer and device.modelName
        device_path = os.path.join(os.getcwd(), "%s %s" % (device.manufacturer, device.modelName))
        commands = gateway.debug_commands()
        with open(device_path, "w") as f:
            f.write(json.dumps({
                "router_address": self.gateway_address,
                "client_address": self.lan_address,
                "port": gateway.port,
                "gateway_dict": gateway.gateway_descriptor(),
                'expected_devices': [
                    {
                        'cache_control': 'max-age=1800',
                        'location': gateway.location,
                        'server': gateway.server,
                        'st': gateway.urn,
                        'usn': gateway.usn
                    }
                ],
                'commands': commands,
                # 'ssdp': u.sspd_factory.get_ssdp_packet_replay(),
                # 'scpd': gateway.requester.dump_packets(),
                'soap': commands_test_case
            }, default=_encode, indent=2).replace(external_ip, "1.2.3.4"))
        return "Generated test data! -> %s" % device_path

    @classmethod
    def run_cli(cls, method, lan_address: str = '', gateway_address: str = '', timeout: int = 60,
                          service: str = UPNP_ORG_IGD, interface_name: str = 'default',
                kwargs: dict = None):
        kwargs = kwargs or {}

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        fut = asyncio.Future()

        async def wrapper():
            if method == 'm_search':
                fn = lambda *_a, **_kw: cls.m_search(lan_address, gateway_address, timeout, service, interface_name)
            else:
                u = await cls.discover(
                    lan_address, gateway_address, timeout, service, interface_name
                )
                if hasattr(u, method) and hasattr(getattr(u, method), "_cli"):
                    fn = getattr(u, method)
                else:
                    fut.set_exception(UPnPError("\"%s\" is not a recognized command" % method))
                    return
            try:
                result = await fn(**kwargs)
                fut.set_result(result)
            except UPnPError as err:
                fut.set_exception(err)
            except Exception as err:
                log.exception("uncaught error")
                fut.set_exception(UPnPError("uncaught error: %s" % str(err)))

        asyncio.run(wrapper())
        try:
            result = fut.result()
        except UPnPError as err:
            print("error: %s" % str(err))
            return

        if isinstance(result, (list, tuple, dict)):
            print(json.dumps(result, indent=2, default=_encode))
        else:
            print(result)
